/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla XForms support.
 *
 * The Initial Developer of the Original Code is
 * IBM Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Aaron Reed <aaronr@us.ibm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsIXTFXMLVisual.h"
#include "nsIXMLContentBuilder.h"
#include "nsIDOMEvent.h"
#include "nsIDOMEventTarget.h"
#include "nsIDOMEventListener.h"
#include "nsIDOM3Node.h"
#include "nsIDOMElement.h"
#include "nsMemory.h"
#include "nsCOMPtr.h"
#include "nsString.h"
#include "nsIXTFXMLVisualWrapper.h"
#include "nsIDOMDocument.h"
#include "nsIDOMHTMLInputElement.h"
#include "nsIDOMHTMLElement.h"
#include "nsIDOMText.h"
#include "nsUnicharUtils.h"
#include "WidgetAtoms.h"
#include "WidgetElement.h"

#include "stdlib.h"

#include "nsComponentManagerUtils.h"  //for do_QueryInterface


static const nsIID sScriptingIIDs[] = {
  NS_IDOMELEMENT_IID,
  NS_IDOMEVENTTARGET_IID,
  NS_IDOM3NODE_IID
};

class WidgetSSNElement : public WidgetElement,
                         public nsIXTFXMLVisual,
                         public nsIDOMEventListener
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_NSIXTFXMLVISUAL
  NS_DECL_NSIXTFVISUAL
  NS_DECL_NSIXTFELEMENT
  NS_DECL_NSIDOMEVENTLISTENER

  virtual NS_HIDDEN_(void) Refresh();
  virtual NS_HIDDEN_(void) ValidateField( nsIDOMNode *field);

private:
  nsCOMPtr<nsIXTFXMLVisualWrapper> mWrapper;
  nsCOMPtr<nsIDOMHTMLElement> mContent;
  nsCOMPtr<nsIDOMHTMLInputElement> mField1;
  nsCOMPtr<nsIDOMHTMLInputElement> mField2;
  nsCOMPtr<nsIDOMHTMLInputElement> mField3;
  nsString mWarningColor;
  PRBool   mChanged;
};

NS_IMPL_ISUPPORTS2(WidgetSSNElement, nsIXTFXMLVisual, nsIXTFElement)

// nsIXTFXMLVisual

// This method gets called when layout encounters our "ssn" tag.  Since we want
//   the "ssn" tag to be represented in the document as 3 HTML input fields
//   seperated by "-"'s, then this is where we build that representation.

#if 0 // both of these OnCreated functions work.  Here are two ways to do
      //   the same thing.

NS_IMETHODIMP
WidgetSSNElement::OnCreated(nsIXTFXMLVisualWrapper *aWrapper)
{
  mWrapper = aWrapper;
  mChanged = PR_FALSE;  // initialize to false so that we don't mark fields
                        //   fields as being invalid on initialization

  nsCOMPtr<nsIDOMElement> node;
  mWrapper->GetElementNode(getter_AddRefs(node));
  nsCOMPtr<nsIDOMDocument> domDoc;
  node->GetOwnerDocument(getter_AddRefs(domDoc));

  nsAutoString warning;
  WidgetAtoms::warningColor->ToString( warning );
  node->GetAttribute( warning, mWarningColor );

  // We'll group everything into a HTML span.
  nsCOMPtr<nsIDOMElement> spanElement;
  domDoc->CreateElementNS(NS_LITERAL_STRING("http://www.w3.org/1999/xhtml"),
                          NS_LITERAL_STRING("span"),
                          getter_AddRefs(spanElement));

  if( spanElement )
  {
    // Create the HTML elements that we need.  Set the maxlength attributes on
    //   each input field to help the user avoid mistakes when filling out
    //   the "ssn" control.

    nsCOMPtr<nsIDOMElement> inputElement;
    nsCOMPtr<nsIDOMNode> spanNode = do_QueryInterface(spanElement);
    nsCOMPtr<nsIDOMNode> resultNode;
    mContent = do_QueryInterface(spanElement);
    domDoc->CreateElementNS(NS_LITERAL_STRING("http://www.w3.org/1999/xhtml"),
                            NS_LITERAL_STRING("input"),
                            getter_AddRefs(inputElement));
   
    if( inputElement )
    {
      mField1 = do_QueryInterface(inputElement);
      mField1->SetAttribute( NS_LITERAL_STRING("maxlength"), NS_LITERAL_STRING("3") );
      mField1->SetAttribute( NS_LITERAL_STRING("size"), NS_LITERAL_STRING("3") );

      nsCOMPtr<nsIDOMNode> field1 = do_QueryInterface(mField1);
      spanNode->AppendChild( field1, getter_AddRefs(resultNode) );
      nsCOMPtr<nsIDOMEventTarget> evtTarget = do_QueryInterface(field1);
      evtTarget->AddEventListener(NS_LITERAL_STRING("change"), this, PR_FALSE);

      // this is where the fields get seperated by the dashes ("-")
      nsCOMPtr<nsIDOMText> textNode;
      domDoc->CreateTextNode(NS_LITERAL_STRING(" - "), getter_AddRefs(textNode));
      spanNode->AppendChild( textNode, getter_AddRefs(resultNode) );

    }
   
    domDoc->CreateElementNS(NS_LITERAL_STRING("http://www.w3.org/1999/xhtml"),
                            NS_LITERAL_STRING("input"),
                            getter_AddRefs(inputElement));
   
    mField2 = do_QueryInterface(inputElement);
   
    if( inputElement )
    {
      mField2 = do_QueryInterface(inputElement);
      mField2->SetAttribute( NS_LITERAL_STRING("maxlength"), NS_LITERAL_STRING("2") );
      mField2->SetAttribute( NS_LITERAL_STRING("size"), NS_LITERAL_STRING("2") );

      nsCOMPtr<nsIDOMNode> field2 = do_QueryInterface(mField2);
      spanNode->AppendChild( field2, getter_AddRefs(resultNode) );
      nsCOMPtr<nsIDOMEventTarget> evtTarget = do_QueryInterface(field2);
      evtTarget->AddEventListener(NS_LITERAL_STRING("change"), this, PR_FALSE);

      nsCOMPtr<nsIDOMText> textNode;
      domDoc->CreateTextNode(NS_LITERAL_STRING(" - "), getter_AddRefs(textNode));
      spanNode->AppendChild( textNode, getter_AddRefs(resultNode) );
    }
   
    domDoc->CreateElementNS(NS_LITERAL_STRING("http://www.w3.org/1999/xhtml"),
                            NS_LITERAL_STRING("input"),
                            getter_AddRefs(inputElement));
   
    mField3 = do_QueryInterface(inputElement);
   
    if( inputElement )
    {
      mField3 = do_QueryInterface(inputElement);
      mField3->SetAttribute( NS_LITERAL_STRING("maxlength"), NS_LITERAL_STRING("4") );
      mField3->SetAttribute( NS_LITERAL_STRING("size"), NS_LITERAL_STRING("4") );

      nsCOMPtr<nsIDOMNode> field3 = do_QueryInterface(mField3);
      spanNode->AppendChild( field3, getter_AddRefs(resultNode) );
      nsCOMPtr<nsIDOMEventTarget> evtTarget = do_QueryInterface(field3);
      evtTarget->AddEventListener(NS_LITERAL_STRING("change"), this, PR_FALSE);
    }
  }

  return NS_OK;
}
#else //take a stab at it using content builder

NS_IMETHODIMP
WidgetSSNElement::OnCreated(nsIXTFXMLVisualWrapper *aWrapper)
{
  nsresult rv;
  mWrapper = aWrapper;
  mChanged = PR_FALSE;  // initialize to false so that we don't mark fields
                        //   fields as being invalid on initialization

  nsCOMPtr<nsIDOMElement> node;
  mWrapper->GetElementNode(getter_AddRefs(node));
  nsCOMPtr<nsIDOMDocument> domDoc;
  node->GetOwnerDocument(getter_AddRefs(domDoc));

  // doesn't look like querying attrs on the node will work during OnCreated
  nsAutoString warning;
  WidgetAtoms::warningColor->ToString( warning );
  node->GetAttribute( warning, mWarningColor );

  nsCOMPtr<nsIXMLContentBuilder> builder = 
               do_CreateInstance("@mozilla.org/xtf/xml-contentbuilder;1", &rv);

  if( NS_FAILED(rv) )
  {
    NS_ERROR("could not create XTF content builder");
    return NS_ERROR_FAILURE;
  }

  NS_ASSERTION( builder, "could not create XTF content builder" );

  nsAutoString name_space(NS_LITERAL_STRING("http://www.w3.org/1999/xhtml"));
  builder->SetElementNamespace( name_space );
  // We'll group everything into a HTML span.
  builder->BeginElement(NS_LITERAL_STRING("span"));
    nsCOMPtr<nsIDOMElement> currEle;
    builder->GetCurrent( getter_AddRefs(currEle) );
    mContent = do_QueryInterface(currEle);

    builder->BeginElement(NS_LITERAL_STRING("input"));
      builder->Attrib( NS_LITERAL_STRING("maxlength"), NS_LITERAL_STRING("3") );
      builder->Attrib( NS_LITERAL_STRING("size"), NS_LITERAL_STRING("3") );
      builder->GetCurrent( getter_AddRefs(currEle) );
      mField1 = do_QueryInterface(currEle);
    builder->EndElement();

    builder->TextNode( NS_LITERAL_STRING("-") );

    builder->BeginElement(NS_LITERAL_STRING("input"));
      builder->Attrib( NS_LITERAL_STRING("maxlength"), NS_LITERAL_STRING("2") );
      builder->Attrib( NS_LITERAL_STRING("size"), NS_LITERAL_STRING("2") );
      builder->GetCurrent( getter_AddRefs(currEle) );
      mField2 = do_QueryInterface(currEle);
    builder->EndElement();

    builder->TextNode( NS_LITERAL_STRING("-") );

    builder->BeginElement(NS_LITERAL_STRING("input"));
      builder->Attrib( NS_LITERAL_STRING("maxlength"), NS_LITERAL_STRING("4") );
      builder->Attrib( NS_LITERAL_STRING("size"), NS_LITERAL_STRING("4") );
      builder->GetCurrent( getter_AddRefs(currEle) );
      mField3 = do_QueryInterface(currEle);
    builder->EndElement();

  builder->EndElement();

  nsCOMPtr<nsIDOMEventTarget> evtTarget = do_QueryInterface(mField1);
  evtTarget->AddEventListener(NS_LITERAL_STRING("change"), this, PR_FALSE);
  evtTarget = do_QueryInterface(mField2);
  evtTarget->AddEventListener(NS_LITERAL_STRING("change"), this, PR_FALSE);
  evtTarget = do_QueryInterface(mField3);
  evtTarget->AddEventListener(NS_LITERAL_STRING("change"), this, PR_FALSE);

  mWrapper->SetNotificationMask( nsIXTFElement::NOTIFY_ATTRIBUTE_SET |     
                                 nsIXTFElement::NOTIFY_ATTRIBUTE_REMOVED ); 
               

  return NS_OK;
}

#endif /* if 0 */

NS_IMETHODIMP
WidgetSSNElement::GetVisualContent(nsIDOMElement **aElement)
{
  NS_ADDREF(*aElement = mContent);
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::GetInsertionPoint(nsIDOMElement * *aInsertionPoint)
{
  return NS_OK;
}

NS_IMETHODIMP 
WidgetSSNElement::GetApplyDocumentStyleSheets(PRBool *aApplyDocumentStyleSheets)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::DidLayout(void)
{
  return NS_OK;
}

#if 0
NS_IMETHODIMP
WidgetSSNElement::GetDisplayType(PRUint32 *aDisplayType)
{
  // let these be inline controls.  If this is a block control, then every HTML
  //   control that we add will be on seperate lines in the presentation.  We
  //   want them all on the same line.
  *aDisplayType = nsIXTFXMLVisual::DISPLAY_INLINE;
  return NS_OK;
}
#endif

// nsIXTFElement

NS_IMETHODIMP
WidgetSSNElement::OnDestroyed()
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::GetElementType(PRUint32 *aType)
{
  *aType = ELEMENT_TYPE_XML_VISUAL;
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::GetIsAttributeHandler(PRBool *aIsHandler)
{
  *aIsHandler = PR_FALSE;
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::GetScriptingInterfaces(PRUint32 *aCount, nsIID ***aArray)
{
  return CloneScriptingInterfaces(sScriptingIIDs,
                                  NS_ARRAY_LENGTH(sScriptingIIDs),
                                  aCount, aArray);
}

#if 0
NS_IMETHODIMP
WidgetSSNElement::GetNotificationMask(PRUint32 *aMask)
{
  *aMask = nsIXTFElement::NOTIFY_ATTRIBUTE_SET;
  return NS_OK;
}
#endif

NS_IMETHODIMP
WidgetSSNElement::WillChangeDocument(nsIDOMDocument *aNewDocument)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::DocumentChanged(nsIDOMDocument *aNewDocument)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::WillChangeParent(nsIDOMElement *aNewParent)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::ParentChanged(nsIDOMElement *aNewParent)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::WillInsertChild(nsIDOMNode *aChild, PRUint32 aIndex)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::ChildInserted(nsIDOMNode *aChild, PRUint32 aIndex)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::WillAppendChild(nsIDOMNode *aChild)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::ChildAppended(nsIDOMNode *aChild)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::WillRemoveChild(PRUint32 aIndex)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::ChildRemoved(PRUint32 aIndex)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::WillSetAttribute(nsIAtom *aName, const nsAString &aValue)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::AttributeSet(nsIAtom *aName, const nsAString &aValue)
{
  // AttributeSet is called during initialization of the document after 
  //   OnCreated is called.  So if you call GetAttribute during OnCreated
  //   the attrs on the element won't be found.

  // if the attr that changed isn't the warning color attribute then ignore.
  //   Also ignore if the user hasn't tried to change the values in the fields
  //   yet.  
  if ((aName != WidgetAtoms::warningColor))
    return NS_OK;

  if( !mWarningColor.Equals( aValue, nsCaseInsensitiveStringComparator() ) )
  {
    mWarningColor.Assign( aValue );
    if( !mChanged )
    {
      Refresh();
    }
  }
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::WillRemoveAttribute(nsIAtom *aName)
{
  return NS_OK;
}

NS_IMETHODIMP
WidgetSSNElement::AttributeRemoved(nsIAtom *aName)
{
  if( aName == WidgetAtoms::warningColor )
  {
    mWarningColor.Assign(NS_LITERAL_STRING(""));
    Refresh();
  }
  return NS_OK;
}


NS_IMETHODIMP
WidgetSSNElement::DoneAddingChildren()
{
  return NS_OK;
}

void
WidgetSSNElement::Refresh()
{
  // Refresh() is called when the "warningColor" attribute is changed.  We
  //   want to re-validate the fields and if they are still invalid, then this
  //   is when they'll pick up the new color.

  // if no field has changed, yet, no sense validating them.
  if (!mChanged)
    return;

  // warningColor attribute may have changed.  Run validation again to pick
  //   up the difference.
  nsCOMPtr<nsIDOMNode> node = do_QueryInterface(mField1);
  ValidateField( node );

  node = do_QueryInterface(mField2);
  ValidateField( node );

  node = do_QueryInterface(mField3);
  ValidateField( node );

}

void
WidgetSSNElement::ValidateField(nsIDOMNode *aField)
{
  // ValidateField() is called when either the "warningColor" attribute has
  //   been changed or the user has changed the value in a field and we need
  //   to see if we consider the new value invalid or not.  The value needs
  //   to be comprised of digits in the range of 0-9 and contain the full
  //   number of those digits that we require (maxLength value on the input
  //   field).

  nsCOMPtr<nsIDOMHTMLInputElement> field = do_QueryInterface( aField );

  PRBool setWarningBackground = PR_FALSE;
  PRInt32 iMaxLen;
  nsAutoString fieldValue;
  field->GetMaxLength( &iMaxLen );
  field->GetValue( fieldValue );


  if( iMaxLen > 0 )
  {
    if( fieldValue.Length() < iMaxLen )
    {
      // not enough digits in the field
      setWarningBackground = PR_TRUE;
    }
    else
    {
      nsAString::const_iterator start, end;
      fieldValue.BeginReading( start );
      fieldValue.EndReading( end );
  
      while( start != end )
      {
        if( !(isdigit(*start)) )
        {
          // found a non-decimal character
          setWarningBackground = PR_TRUE;
          break;
        }
        ++start;
      }
    }
  }

  nsCOMPtr<nsIDOMElement> domField = do_QueryInterface( field );
  if( setWarningBackground == PR_TRUE )
  {
    if( !mWarningColor.IsEmpty() )
    {
      // We know that we need to indicate that there is an invalid value in the
      //   field and the document author told us what color to use to catch
      //   the user's attention.  Make it so!

      // build style string to set the background color of the control to the
      //   value of the warning color attribute set on the WidgetSSNElement.
      //   No one seems to know how to do this properly, so I'll do it like
      //   a JS author would do it...set the style attribute on the control.

      nsAutoString backgroundStyle = NS_LITERAL_STRING("background:") + mWarningColor;

      domField->SetAttribute( NS_LITERAL_STRING("style"), backgroundStyle );
    }
  }
  else
  {
    // Well, we don't need to have a warning color showing.  By setting the attr
    //   to a null string, that will reset it to the default style.
    domField->SetAttribute( NS_LITERAL_STRING("style"), NS_LITERAL_STRING("") );
  }

}


NS_IMETHODIMP
WidgetSSNElement::HandleEvent(nsIDOMEvent* aChangeEvent)
{
  // This is called whenever a DOM change event occurs.  This listener is
  //   added to each field.

  nsCOMPtr<nsIDOMEventTarget> target;
  nsresult res = aChangeEvent->GetTarget( getter_AddRefs(target) );

  nsCOMPtr<nsIDOMNode> node = do_QueryInterface( target );
  mChanged = PR_TRUE;
  ValidateField( node );

  return NS_OK;
}


NS_HIDDEN_(nsresult)
NS_NewWidgetSSNElement(nsIXTFElement **aResult)
{
  *aResult = new WidgetSSNElement();
  if (!*aResult)
    return NS_ERROR_OUT_OF_MEMORY;

  NS_ADDREF(*aResult);
  return NS_OK;
}
